//
//	*** Modbus RTU server - Function definitions - Version 1 ***
//
//	SoftwareSerial instantiated in heap
//
#include <Arduino.h>
#include <stdint.h>
#include "ModbusServer.h"
//{
//Serial.println("ModbusRtuSerialServer1 constructor");
//}

ModbusServer::ModbusServer(byte id, byte rxpin, byte txpin, uint32_t baud,
		uint16_t nbrHregs, uint16_t nbrCoils, int * hregs, byte * coils) :
		SoftwareSerial(rxpin, txpin)
{
	_device = id;
	_baud = baud;
	_frameDelay = 24000 / _baud;
	//_frameDelay = 100000/_baud;
	_nbrCoils = nbrCoils;
	_nbrHregs = nbrHregs;
	_coils = coils;
	_hregs = hregs;
	_len = 0;
	_errcnt = 0;
	_crc = 0;

	//Initialize Serial link
	begin(_baud);
	flush();
}
/*
 Generates the crc for the current message in buffer.
 */
void ModbusServer::calcCrc(void)
{
	uint16_t crc = 0XFFFF;
	byte i, lsb, msgLen, *msgPtr;
	msgLen = _len - 2;
	msgPtr = _msg;
	while (msgLen--)
	{
		crc = crc ^ *msgPtr++;
		for (i = 0; i <= 7; i++)
		{
			lsb = crc & 1;
			crc = crc >> 1;
			if (lsb)
				crc = crc ^ 0xA001;
		}

	}
	_crc = ((crc << 8) | crc >> 8) & 0xFFFF;
}

/*
 Checks the UART for query data
 */
void ModbusServer::checkSerial(void)
{
	//while there is more data in the UART than when last checked
	while (available() > _len)
	{
		_len = available();  	      //update the incoming query message length
		delay(_frameDelay);	//Wait for 3 byte widths of data (SOM/EOM)
		//Serial.print(_len);
	}
}

/*
 Copies the contents of the UART to a buffer
 */
void ModbusServer::serialRx(void)
{
	byte i;
	//copy the query byte for byte to the new buffer
	for (i = 0; i < _len; i++)
	{
		_msg[i] = read();
		//Serial.print(_msg[i],HEX);
	}
	//Serial.println();
}

/*
 Generates a query reply message for Digital In/Out status update queries.
 */

void ModbusServer::getDigitalStatus(byte funcType, uint16_t startreg,
		uint16_t numregs)
{
	uint16_t first;
	uint16_t bytecnt = 0;
	byte nbytes = 0;

	first = startreg / 8;
	_len = numregs / 8;	//determine the message length
	if (numregs % 8)		//add one if partial byte requested
		_len++;
	if (_len > _nbrCoils)
		_len = _nbrCoils;   //Limit access to table bounds
	nbytes = _len;		//number of bytes to return
	_len += 5;//allow room for the Device ID byte, Function type byte, data byte count byte, and crc word
	_msg[0] = _device;	//write the slave device ID
	_msg[1] = funcType;	//write the function type
	_msg[2] = _len - 5;	//set the data byte count

	//Loop to get all bytes requested
	//Serial.print("Booleans: ");

	while (nbytes--)
	{
		_msg[3 + bytecnt] = _coils[first + bytecnt];
		//Serial.print(_msg[3 + bytecnt],HEX);
		//Serial.print(",");
		bytecnt++;
	}
	//Serial.println();
	//generate the crc for the query reply and append it
	this->calcCrc();
	_msg[_len - 2] = _crc >> 8;
	_msg[_len - 1] = _crc & 0xFF;
}

void ModbusServer::getAnalogStatus(byte funcType, uint16_t startreg,
		uint16_t numregs)
{
	uint16_t val;
	uint16_t i = 0;

	//calculate the query reply message length,for each register queried add 2 bytes
	if (numregs > _nbrHregs)
		numregs = _nbrHregs; //Limit access to table boundaries
	_len = numregs * 2;
	_msg[0] = _device; 		//write the device ID
	_msg[1] = funcType;		//write the function type
	_msg[2] = _len;			//set the data byte count
	_len += 5;//allow room for the Device ID byte, Function type byte, data byte count byte, and crc word

	//for each register queried
	//Serial.print("Analogs: ");
	while (numregs--)
	{
		//retrieve the value from the register bank for the current register
		val = _hregs[startreg + i];	//Get value from analog data table
		_msg[3 + i * 2] = val >> 8;	//write the high byte of the register value
		_msg[4 + i * 2] = val & 0xFF;//write the low byte of the register value
		//Serial.print(_msg[3 + i],HEX);
		//Serial.print(",");

		i++;	//increment the register
	}
	//Serial.println();
	this->calcCrc();	//generate the crc for the query reply and append it
	_msg[_len - 2] = _crc >> 8;
	_msg[_len - 1] = _crc & 0xFF;
}

void ModbusServer::setStatus(byte funcType, uint16_t reg, uint16_t val)
{
	//Set the query response message length
	//Device ID byte, Function byte, Register byte, Value byte, CRC word
	_len = 8;
	_msg[0] = _device;	//write the device ID
	//if the function type is a digital write

	if (funcType == WRITE_DO)
	{
		if (reg > (_nbrCoils - 1))
			reg = _nbrCoils - 1; //Limit access to table bounds
		word index = reg / 8;		//byte for requested coil
		byte mask = 1 << (reg % 8);	//bit for requested coil
		_coils[index] = val ? _coils[index] | mask : _coils[index] & ~mask;
		_msg[1] = WRITE_DO;	//write the function type to the response message

		//printf("MBS set Status: code= %d, reg= %d, val=%2X\n",funcType, reg, val);
		//printf("DO inx= %d, mask: %2x\n", index,mask);
	}
	else
	{
		if (reg > _nbrHregs)
			reg = _nbrHregs; //Limit access to table bounds
		_hregs[reg] = val;	//Set value into register
		_msg[1] = WRITE_AO;	//write the function type of the response message
	}
	_msg[2] = reg >> 8;		//write the register number high byte value
	_msg[3] = reg & 0xFF;	//write the register number low byte value
	_msg[4] = val >> 8;		//write the control value's high byte
	_msg[5] = val & 0xFF;	//write the control value's low byte

	this->calcCrc();	//calculate the crc for the query reply and append it.
	_msg[_len - 2] = _crc >> 8;
	_msg[_len - 1] = _crc & 0xFF;
}

void ModbusServer::run(void)
{
	byte funcType;
	uint16_t field1;
	uint16_t field2;
	//initialize mesasge length
	_len = 0;

	//check for data in the recieve buffer
	this->checkSerial();

	//if there is nothing in the receive buffer, bail.
	if (_len == 0)
	{
		return;
	}
	this->serialRx();	//retrieve the query message from the serial uart

	//printf("MBrx: ");
	//for (int i=0; i<_len;i++)
	//{
	//   printf("%2.2X, ",_msg[i]);
	//}
	//printf("\n");

	if ((_msg[0] != 0xFF) && (_msg[0] != _device))
		return;	//Ignore command id is not actual device id or 255 (broadcast)

	this->calcCrc();//calculate the checksum of the query message minus the checksum it came with.
	if (_crc != (uint16_t) ((_msg[_len - 2] << 8) + _msg[_len - 1]))//if the checksum does not match, ignore the message
	{
		printf("MB CRC error: %ld\n", ++_errcnt);
		return;
	}
	funcType = _msg[1];	//copy the function type from the incoming query
	field1 = (_msg[2] << 8) | _msg[3];//copy field 1 from the incoming query (start address)
	field1 &= 0xFFF;	//Remove standard modbus offset (0X1000 to 0x7000)
	field2 = (_msg[4] << 8) | _msg[5];//copy field 2 from the incoming query (number of items)
	_len = 0;			//initialize response byte counter

	//generate query response based on function type
	switch (funcType)
	{
	case READ_DI:
		this->getDigitalStatus(funcType, field1, field2);
		break;
	case READ_DO:
		this->getDigitalStatus(funcType, field1, field2);
		break;
	case READ_AI:
		this->getAnalogStatus(funcType, field1, field2);
		break;
	case READ_AO:
		this->getAnalogStatus(funcType, field1, field2);
		break;
	case WRITE_DO:
		this->setStatus(funcType, field1, field2);
		break;
	case WRITE_AO:
		this->setStatus(funcType, field1, field2);
		break;
	default:
		return;
		break;
	}

	//if a reply was generated
	if (_len)
	{
		int i;
		for (i = 0; i < _len; i++)
			write(_msg[i]);
		_len = 0;
	}
}
